#/******************************************************************************
#                                ZLib : 11
#                    supporting functions for scripts
#*******************************************************************************
#
#   Want to say thanks?  Send me a bat! (Or bat-related item)
#
#   For a list of included functions, check out
#   http://kolmafia.us/showthread.php?t=2072
#
#******************************************************************************/

// set your password for the Wikified Online Script Settings Manager (WOSSMAN)
// (leave blank to disable WOSSMAN)
string wossman_pwd = "";

//                             NUMBER ROUTINES

float minmax(float a, float min, float max) {
   if (a < min) return min;
   if (a > max) return max;
   return a;
}
void set_avg(float toadd, string whichprop) {
   string initv = get_property(whichprop);
   if (initv == "") { set_property(whichprop,toadd+":1"); return; }
   float a = to_float(substring(initv,0,index_of(initv,":")));
   float b = to_float(substring(initv,index_of(initv,":"),length(initv)));
   a = ((a * b)+toadd) / (b+1);
   b = b + 1;
   set_property(whichprop,a+":"+b);
}
float get_avg(string whichprop) {
   string initv = get_property(whichprop);
   if (initv == "") return 0;
   return to_float(substring(initv,0,index_of(initv,":")));
}

// Evaluates expressions in the format used by variable modifiers:
//
// No spaces allowed, except as part of a zone/location name.
// + - * / ( ) have their usual mathematical meaning and precedence.
// ^ is exponentiation, with the highest precedence.
// Math functions: ceil(x) floor(x) sqrt(x) min(x,y) max(x,y)
// Text functions (can use a maximum of ONE!):
//   loc(text), zone(text) - 1 if the current adventure location or zone contains the specified text, 0 elsewise.
//   fam(text) - 1 if the player's familiar type contains the text, else 0.
//   pref(text) - must be used on preferences with a float value ONLY - merely retrieving an integer pref will corrupt it!
// Internally-used variables (upper-case letters)
//	D - drunkenness
//	F - fullness
//	G - Grimace darkness (0..5)
//	L - player level
//	M - total moonlight (0..9)
//	S - spleenness
//	X - gender (-1=male, 1=female)
// User-defined variables (provided in "values") must begin with lowercase letters/underscores
float eval(string expr, float[string] values) {
   buffer b;
   matcher m = create_matcher("\\b[a-z_][a-zA-Z0-9_]*\\b", expr);
   while (m.find()) {
      string var = m.group(0);
      if (values contains var) m.append_replacement(b, values[var].to_string());
   }
   m.append_tail(b);
//   print("Evaluating '"+b.to_string()+"'...");
   return modifier_eval(b.to_string());
}

//                             STRING ROUTINES

// returns the string between start and end in source
string excise(string source, string start, string end) {
   if (!source.contains_text(start)) return "";
   string result = substring(source,index_of(source,start)+length(start));
   if (!result.contains_text(end)) return "";
   return substring(result,0,index_of(result,end));
}

// wrapper for (print(error message,red); return false;) often better than abort()
boolean error(string message) { print(message,"red"); return false; }

// mafia lacks this
element to_element(string s) {
   switch (s) {
      case "hot":
      case "heat": return $element[hot];
      case "cold": return $element[cold];
      case "spooky": return $element[spooky];
      case "sleaze": return $element[sleaze];
      case "stench": return $element[stench];
   }
   return $element[none];
}

//                             SCRIPT ROUTINES

// checks script version once daily
void check_version(string soft, string prop, string thisver, int thread) { int w;
   switch (get_property("_version_"+prop)) {
      case thisver: return;
      case "": print("Checking for updates (running "+soft+" ver. "+thisver+")...");
         matcher find_ver = create_matcher("<b>"+soft+" (.+?)</b>",visit_url("http://kolmafia.us/showthread.php?t="+thread)); w=20;
         if (!find_ver.find()) { print("Unable to load current version info."); return; }
         set_property("_version_"+prop,find_ver.group(1));
         if (find_ver.group(1) == thisver) { print("You have a current version of "+soft+"."); return; }
      default:
         print_html("<big><font color=red><b>New Version of "+soft+" Available: "+get_property("_version_"+prop)+"</b></font></big>");
         print("Upgrade "+soft+" from "+thisver+" to "+get_property("_version_"+prop)+" here: http://kolmafia.us/showthread.php?t="+thread,"blue");wait(3+w);
   }
}

//                                 WOSSMAN

string[string] vars;
file_to_map("vars_"+replace_string(my_name()," ","_")+".txt",vars);

// updates vars map
boolean updatevars(string url) {
   if (url == "") url = "action=map";
//   if (wossman_pwd != "")
//      file_to_map("http://zachbardon.com/mafiatools/wossman.php?id="+my_id()+"&pass="+url_encode(wossman_pwd)+"&"+url,vars);
   return map_to_file(vars,"vars_"+replace_string(my_name()," ","_")+".txt");
}
void setvar(string varname,string dfault,string type) {
   if (vars contains varname) return;
   vars[varname] = dfault;
   updatevars("action=entry&varname="+url_encode(varname)+"&value="+url_encode(dfault)+"&type="+type);
}
void setvar(string varname,string dfault) {  setvar(varname,dfault,"string");  }
void setvar(string varname,boolean dfault) {  setvar(varname,to_string(dfault),"boolean");  }
void setvar(string varname,int dfault) {  setvar(varname,to_string(dfault),"int");  }
void setvar(string varname,float dfault) {  setvar(varname,to_string(dfault),"float");  }
void setvar(string varname,item dfault) {  setvar(varname,to_string(dfault),"item");  }
void setvar(string varname,location dfault) {  setvar(varname,to_string(dfault),"location");  }
void setvar(string varname,element dfault) {  setvar(varname,to_string(dfault),"element");  }
void setvar(string varname,familiar dfault) {  setvar(varname,to_string(dfault),"familiar");  }
void setvar(string varname,skill dfault) {  setvar(varname,to_string(dfault),"skill");  }
void setvar(string varname,effect dfault) {  setvar(varname,to_string(dfault),"effect");  }
void setvar(string varname,stat dfault) {  setvar(varname,to_string(dfault),"stat");  }
void setvar(string varname,class dfault) {  setvar(varname,to_string(dfault),"class");  }


//                           ADVENTURING ROUTINES

// returns how many of an item you have
int have_item(string tolookup) {
   return item_amount(to_item(tolookup)) + equipped_amount(to_item(tolookup));
}

// gets n (-existing) of cond, either by purchasing, pulling from Hangk's, or
// adventuring at the given location.  also works with choiceadvs
boolean obtain(int n, string cond, location locale) {
   if (cond == "choiceadv") cli_execute("conditions clear; conditions add "+n+" choiceadv");
   else {
      if (retrieve_item(n, to_item(cond))) return true;
      if (!in_hardcore() && storage_amount(to_item(cond)) > 0 && take_storage(n-have_item(cond),to_item(cond)))
         return true;
      cli_execute("conditions clear");
      add_item_condition(n - have_item(cond), to_item(cond));
   }
   set_location(locale);
   if (adventure(my_adventures(), locale)) return error("Out of adventures.");
   if (cond == "choiceadv") return (my_adventures() > 0);
   return (have_item(cond) >= n);
}

// gets (if purchase) and uses n doodads if possible
// otherwise, uses as many as you have up to n
boolean use_upto(int n, item doodad, boolean purchase) {
   if (doodad == $item[deodorant] && item_amount($item[chunk of rock salt]) > item_amount(doodad)) doodad = $item[chunk of rock salt];
   if (item_amount(doodad) >= n || (purchase && retrieve_item(n, doodad))) return use(n, doodad);
   if (item_amount(doodad) == 0) return false;
   return use(item_amount(doodad), doodad);
}

// attempts to resist the given element
boolean resist(element req, boolean reallydoit) {
   print("Trying to resist "+req+"...");
   if (req == $element[none]) return true;
   if (elemental_resistance(req) >= 10) return true;
  // first, try catchall buffs
   if (have_skill($skill[astral shell]) && (
       (!reallydoit && my_mp() >= mp_cost($skill[astral shell])) ||
       (reallydoit && use_skill(1,$skill[astral shell])))) return true;
   if (have_skill($skill[elemental saucesphere]) && (
       (!reallydoit && my_mp() >= mp_cost($skill[elemental saucesphere])) ||
       (reallydoit && use_skill(1,$skill[elemental saucesphere])))) return true;
  // next, try resistance from gear
   int[item] mgear = get_inventory();
   string rtype = to_string(req) + " resistance";
   print("Searching items for "+rtype+"...");
   foreach doodad in mgear {
      if (to_slot(doodad) != $slot[none] && numeric_modifier(doodad,rtype) >= 1.0 && can_equip(doodad)) {
         print("Resistance-granting item found: "+doodad);
         if (reallydoit) {
            if (to_slot(doodad) == $slot[acc1]) equip($slot[acc3],doodad);
              else equip(doodad);
         }
         return true;
      }
   }
  // last, try resistance from the parrot
   if (have_familiar($familiar[exotic parrot])) {
       int necessary_parrot_weight(element which) {
          switch (which) {
             case $element[hot]: return 1;
             case $element[cold]: return 5;
             case $element[spooky]: return 9;
             case $element[stench]: return 13;
             case $element[sleaze]: return 17;
          }
          return 0;
       }
       int possible_parrot_weight() {
          int result = familiar_weight($familiar[exotic parrot]);
          if (familiar_equipped_equipment($familiar[exotic parrot]) == $item[cracker] || item_amount($item[cracker]) > 0)
             result = result + 15;
          return result + numeric_modifier("Familiar Weight") - numeric_modifier(equipped_item($slot[familiar]),"Familiar Weight");
       }
       if (possible_parrot_weight() >= necessary_parrot_weight(req)) {
          print("Your parrot is able to resist "+req+".");
          if (reallydoit) {
             use_familiar($familiar[exotic parrot]);
             if (familiar_equipped_equipment($familiar[exotic parrot]) != $item[cracker] && item_amount($item[cracker]) > 0)
                equip($item[cracker]);
          }
          return true;
       }
   }
   return error("Unable to resist "+req+"!");
}

// returns the value of your buffed defense stat, factoring in Hero of the Half-Shell
int my_defstat() {
   if (have_skill($skill[hero of the half-shell]) && item_type(equipped_item($slot[offhand])) == "shield")
      return max(my_buffedstat($stat[muscle]),my_buffedstat($stat[moxie]));
   return my_buffedstat($stat[moxie]);
}

// returns safe moxie for a given location (includes +ML and MCD)
int get_safemox(location wear) {
   int high = 0;
   foreach i,m in get_monsters(wear)
      high = max(monster_attack(m),high);
   if (high == 0 || high == monster_level_adjustment()) return 0;
   return high + 7 - current_mcd();
}

// if 'automcd' is true, adjusts your MCD for the specified safemox based on your threshold
boolean auto_mcd(int safemox) {
   if (!to_boolean(vars["automcd"]) || my_location() == $location[boss bat's lair] ||
       my_location() == $location[king's chamber] || my_location() == $location[haert of the cyrpt]) return true;
   if ( (!in_muscle_sign() && !in_moxie_sign() && !in_mysticality_sign()) ||
     (in_muscle_sign() && !retrieve_item(1,$item[detuned radio])) ||
     (in_moxie_sign() && available_amount($item[bitchin' meatcar]) == 0) &&
      !contains_text(get_property("unlockedLocations"),"South of the Border") &&
      !contains_text(get_property("unlockedLocations"),"Thugnderdome")   ) {
      print("MCD: unavailable","olive");
      return true;
   }
   if (safemox == 0) {
      print("MCD: Using your 'unknown_ml' value ("+vars["unknown_ml"]+").","olive");
      safemox = to_int(vars["unknown_ml"] + 7);
   }
   int adj = minmax(my_defstat() + to_int(vars["threshold"]) - safemox, 0, 10+in_mysticality_sign().to_int());
   if (current_mcd() == adj) return true;
   else print("MCD: adjusting to "+adj+"...","olive");
   return change_mcd(adj);
}
boolean auto_mcd(location place) { return auto_mcd(get_safemox(place)); }      // automcd for locations
boolean auto_mcd(monster mob) {                                                // automcd for a single monster
   if (monster_attack(mob) == monster_level_adjustment()) return auto_mcd(0);
   return auto_mcd(monster_attack(mob) + 7 - current_mcd());
}

boolean load_fam_map(string fname, string[familiar] dest) {
   string curr = visit_url("http://zachbardon.com/mafiatools/autoupdate.php?f="+fname+"&act=getver");
   file_to_map(fname+".txt",dest);
   if  ((count(dest) == 0) || (curr != "" && get_property("map_"+fname+".txt") != curr)) {
      print("Updating "+fname+".txt from '"+get_property("map_"+fname+".txt")+"'...");
      if (!file_to_map("http://zachbardon.com/mafiatools/autoupdate.php?f="+fname+"&act=getmap",dest) || count(dest) == 0) return false;
      map_to_file(dest,fname+".txt");
      set_property("map_"+fname+".txt",curr);
      print("..."+fname+".txt updated.");
   }
   return true;
}

// returns your heaviest familiar of a given type (currently possible: items, meat, produce, stat, delevel)
familiar best_fam(string ftype) {
   if (to_familiar(vars["is_100_run"]) != $familiar[none]) return to_familiar(vars["is_100_run"]);
   familiar result = $familiar[none];
   string[familiar] fams;
   if (!load_fam_map("bestfamiliars",fams)) return result;
   int bestweight = 0;
   int beststrength = 0;
   int famstrength(string alltypes, string targettype) {
     matcher m = create_matcher("(\\w+)\\s+(\\d+)", alltypes);          // thanks for the regex Jason
     while (m.find()) if (m.group(1) == targettype) return to_int(m.group(2));
     return 0;
   }
   foreach i in fams {
     if (!have_familiar(i) || !contains_text(fams[i],ftype)) continue;
     if (familiar_weight(i) > bestweight || (familiar_weight(i) == bestweight && famstrength(fams[i],ftype) > beststrength)) {
        if (result == $familiar[none]) print("First match for type \""+ftype+"\": "+i);
         else print(i+" is better than "+result+"...");
        result = i;
        bestweight = familiar_weight(i);
        beststrength = famstrength(fams[i],ftype);
     }
   }
   if (result == $familiar[none]) {
      print("No familiar found matching type \""+ftype+"\"");
      return my_familiar();
   }
   return result;
}

boolean send_gift(string to, string message, int meat, int[item] goodies, string insidenote) {
 // parse items into query string
   string itemstring = "";
   int j = 0;
   int[item] extra;
   if (count(goodies) > 0)
   foreach i in goodies {
      if (is_tradeable(i) || is_giftable(i)) {
         j = j+1;
         if (j < 4)
           itemstring = itemstring + "&howmany"+j+"="+goodies[i]+"&whichitem"+j+"="+to_int(i);
         else extra[i] = goodies[i];
      }
   }
   int shipping = 200;
   int pnum = 3;
   if (count(goodies) < 3) {
      shipping = 50*count(goodies);
      pnum = count(goodies);
   }
   if (my_meat() < meat+shipping) return error("Not enough meat to send the package.");
  // send gift
   string url = visit_url("town_sendgift.php?pwd=&towho="+to+"&note="+message+"&insidenote="+insidenote+"&whichpackage="+pnum+"&fromwhere=0&sendmeat="+meat+"&action=Yep."+itemstring);
   if (!contains_text(url,"Package sent.")) return error("The message didn't send for some reason.");
   if (count(extra) > 0) return send_gift(to,message,0,extra,insidenote);
   return true;
}
boolean send_gift(string to, string message, int meat, int[item] goodies) { return send_gift(to,message,meat,goodies,""); }

boolean kmail(string to, string message, int meat, int[item] goodies, string insidenote) {
  // parse items into query strings
   string itemstring = "";
   int j = 0;
   string[int] itemstrings;
   if (count(goodies) > 0)
   foreach i in goodies {
      if (is_tradeable(i) || is_giftable(i)) {
         j = j+1;
         itemstring = itemstring + "&howmany"+j+"="+goodies[i]+"&whichitem"+j+"="+to_int(i);
         if (j > 10) {
            itemstrings[count(itemstrings)] = itemstring;
            itemstring = '';
            j = 0;
         }
      }
   }
   if (itemstring != "") itemstrings[count(itemstrings)] = itemstring;
   if (count(itemstrings) == 0) itemstrings[0] = "";
  // send message(s)
   foreach q in itemstrings {
      string url = visit_url("sendmessage.php?pwd=&action=send&towho="+url_encode(to)+"&message="+url_encode(message)+"&savecopy=on&sendmeat="+meat+itemstrings[q]);
      if (contains_text(url,"That player cannot receive Meat or items"))
        return send_gift(to, message, meat, goodies, insidenote);
      if (!contains_text(url,"Message sent.")) return error("The message didn't send for some reason.");
   }
   return true;
}
boolean kmail(string to, string message, int meat, int[item] goodies) { return kmail(to,message,meat,goodies,""); }
boolean kmail(string to, string message, int meat) { int[item] nothing; return kmail(to,message,meat,nothing,""); }

// NOTE: after running this script, changing these variables here in the script will have no
// effect.  You should instead edit the values in vars_myname.txt, in mafia's data directory.
setvar("automcd",true);
setvar("threshold",4);
setvar("unknown_ml",170);
setvar("is_100_run",$familiar[none]);
setvar("defaultoutfit","current");

check_version("ZLib","zlib",": 11",2072);
// if (wossman_pwd == "" && get_property("wossman_pwd") != "") wossman_pwd = get_property("wossman_pwd");
// else if (wossman_pwd != "") set_property("wossman_pwd",wossman_pwd);
void main(string setval) {
/*
   if (setval == "" || setval == "update") {
      if (wossman_pwd == "") abort("WOSSMAN is disabled.  To enable it, set the wossname_pwd variable in zlib.ash to a password of your choosing.");
      if (updatevars("")) print("Your script settings have been updated from the WOSSMAN.","green");
      return;
   }
*/
   if (setval == "vars") { foreach key,val in vars print(key+" = "+val); return; }
   if (!setval.contains_text(" = ")) abort("Proper syntax is settingname = value");
   string n = substring(setval,0,index_of(setval," = "));
   if (!(vars contains n)) { print("No setting named '"+n+"' exists.","olive"); return; }
   string v = substring(setval,index_of(setval," = ")+3);
   print("Previous value of "+n+": "+vars[n]);
   if (vars[n] == v) return;
   if (n == "threshold") {
      if (v == "up") v = to_string(to_int(vars["threshold"])+1);
      if (v == "down") v = to_string(to_int(vars["threshold"])-1);
   }
   vars[n] = v;
   updatevars("action=entry&varname="+url_encode(n)+"&value="+url_encode(v));
   print("Changed to "+v+".");
}